from typing import Any, List


def pad_array(in_list: List[Any], max_length: int, value: Any = 0) -> List[Any]:
    """Pad a list to the given size with the given padding value.

    # Arguments:
    in_list (List[Any]): List of values to pad to the given length
    max_length (int): The desired length of the array
    value (Any): a value to insert in order to pad the array to max length

    # Returns
    (List[Any]): An array padded to `max_length` size
    """
    while len(in_list) < max_length:
        in_list.append(value)
    return in_list



from typing import Any

from ..envs.complex_simplify import ComplexSimplify
from ..types import MathyEnvDifficulty, MathyEnvProblemArgs
from .mathy_gym_env import MathyGymEnv, safe_register


class GymComplexTerms(MathyGymEnv):
    def __init__(self, difficulty: MathyEnvDifficulty, **kwargs: Any):
        super(GymComplexTerms, self).__init__(
            env_class=ComplexSimplify,
            env_problem_args=MathyEnvProblemArgs(difficulty=difficulty),
            **kwargs
        )


class ComplexTermsEasy(GymComplexTerms):
    def __init__(self, **kwargs: Any):
        super(ComplexTermsEasy, self).__init__(
            difficulty=MathyEnvDifficulty.easy, **kwargs
        )


class ComplexTermsNormal(GymComplexTerms):
    def __init__(self, **kwargs: Any):
        super(ComplexTermsNormal, self).__init__(
            difficulty=MathyEnvDifficulty.normal, **kwargs
        )


class ComplexTermsHard(GymComplexTerms):
    def __init__(self, **kwargs: Any):
        super(ComplexTermsHard, self).__init__(
            difficulty=MathyEnvDifficulty.hard, **kwargs
        )


safe_register(id="mathy-complex-easy-v0", entry_point="mathy_envs.gym:ComplexTermsEasy")
safe_register(
    id="mathy-complex-normal-v0", entry_point="mathy_envs.gym:ComplexTermsNormal"
)
safe_register(id="mathy-complex-hard-v0", entry_point="mathy_envs.gym:ComplexTermsHard")



from typing import Optional

from mathy_core.expressions import MathExpression
from mathy_core.problems import gen_commute_haystack, rand_bool
from mathy_core.rules import DistributiveFactorOutRule
from numpy.random import randint

from .. import time_step
from ..env import MathyEnvProblem
from ..state import MathyEnvState, MathyObservation
from ..types import MathyEnvDifficulty, MathyEnvProblemArgs
from .poly_simplify import PolySimplify


class PolyCommuteLikeTerms(PolySimplify):
    """A Mathy environment for moving like terms near each other to enable
    further simplification.

    This task is intended to test the model's ability to identify like terms
    in a large string of unlike terms and its ability to use the commutative
    swap rule to reorder the expression bringing the like terms close together.
    """

    rule = DistributiveFactorOutRule()

    def get_env_namespace(self) -> str:
        return "mathy.polynomials.commute.like.terms"

    def transition_fn(
        self,
        env_state: MathyEnvState,
        expression: MathExpression,
        features: MathyObservation,
    ) -> Optional[time_step.TimeStep]:
        """If the expression has any nodes that the DistributiveFactorOut rule
        can be applied to, the problem is solved."""
        if self.rule.find_node(expression) is not None:
            return time_step.termination(features, self.get_win_signal(env_state))
        return None

    def max_moves_fn(
        self, problem: MathyEnvProblem, config: MathyEnvProblemArgs
    ) -> int:
        """This task is to move two terms near each other, which requires
        as many actions as there are blocker nodes. The problem complexity
        is a direct measure of this value."""

        return problem.complexity * 4

    def problem_fn(self, params: MathyEnvProblemArgs) -> MathyEnvProblem:
        easy = False
        if params.difficulty == MathyEnvDifficulty.easy:
            blockers = randint(1, 3)
            powers = rand_bool(20)
            text, _ = gen_commute_haystack(
                commute_blockers=blockers,
                min_terms=4,
                max_terms=10,
                easy=easy,
                powers=powers,
            )
        elif params.difficulty == MathyEnvDifficulty.normal:
            blockers = randint(2, 4)
            powers = rand_bool(40)
            text, _ = gen_commute_haystack(
                commute_blockers=blockers,
                min_terms=8,
                max_terms=12,
                easy=easy,
                powers=powers,
            )
        elif params.difficulty == MathyEnvDifficulty.hard:
            blockers = randint(5, 10)
            powers = rand_bool(60)
            text, _ = gen_commute_haystack(
                commute_blockers=blockers,
                min_terms=6,
                max_terms=12,
                easy=easy,
                powers=powers,
            )
        else:
            raise ValueError(f"Unknown difficulty: {params.difficulty}")
        return MathyEnvProblem(text, blockers + 1, self.get_env_namespace())



import random
from typing import Any, Dict, List, Optional, Type

from mathy_core.expressions import MathExpression
from mathy_core.problems import get_rand_term_templates, mathy_term_string
from mathy_core.rule import BaseRule
from mathy_core.rules import (
    AssociativeSwapRule,
    CommutativeSwapRule,
    ConstantsSimplifyRule,
    DistributiveFactorOutRule,
    DistributiveMultiplyRule,
    VariableMultiplyRule,
)
from mathy_core.util import TermEx, get_term_ex, get_terms

from .. import time_step
from ..env import MathyEnvProblem
from ..state import MathyEnvState, MathyEnvStateStep, MathyObservation
from ..types import EnvRewards, MathyEnvDifficulty, MathyEnvProblemArgs
from .poly_simplify import PolySimplify


class PolyHaystackLikeTerms(PolySimplify):
    """Act on any node in the expression that has another term like it
    somewhere else. For example in the problem:

    2x + 8 + 13.2y + z^2 + 5x
    ^^---------------------^^

    Applying any rule to one of those nodes is a win. The idea here is that
    in order to succeed at this task, the model must build a representation
    that can identify like terms in a large expression tree.
    """

    def __init__(self, **kwargs: Any):
        super(PolyHaystackLikeTerms, self).__init__(**kwargs)

    def get_env_namespace(self) -> str:
        return "mathy.polynomials.haystack.like.terms"

    def get_penalizing_actions(self, state: MathyEnvState) -> List[Type[BaseRule]]:
        return [
            CommutativeSwapRule,
            AssociativeSwapRule,
            DistributiveFactorOutRule,
            DistributiveMultiplyRule,
            ConstantsSimplifyRule,
            VariableMultiplyRule,
        ]

    def max_moves_fn(
        self, problem: MathyEnvProblem, config: MathyEnvProblemArgs
    ) -> int:
        return problem.complexity

    def transition_fn(
        self,
        env_state: MathyEnvState,
        expression: MathExpression,
        features: MathyObservation,
    ) -> Optional[time_step.TimeStep]:
        """If all like terms are siblings."""
        agent = env_state.agent
        if len(agent.history) == 0:
            return None
        # History gets pushed before this fn, so history[-1] is the current state,
        # and history[-2] is the previous state. Find the previous state node we
        # acted on, and compare to that.
        curr_timestep: MathyEnvStateStep = agent.history[-1]
        last_timestep: MathyEnvStateStep = agent.history[-2]
        expression = self.parser.parse(last_timestep.raw)
        action_node = self.get_token_at_index(expression, curr_timestep.action[1])
        touched_term = get_term_ex(action_node)

        term_nodes = get_terms(expression)
        # We have the token_index of the term that was acted on, now we have to see
        # if that term has any like siblings (not itself). We do this by ignoring the
        # term with a matching r_index to the node the agent acted on.
        #
        # find_nodes updates the `r_index` value on each node which is the token index
        BaseRule().find_nodes(expression)

        like_counts: Dict[str, int] = {}
        all_indices: Dict[str, List[int]] = {}
        max_index = 0
        for term_node in term_nodes:
            assert term_node is not None and term_node.r_index is not None
            max_index = max(max_index, term_node.r_index)
            ex: Optional[TermEx] = get_term_ex(term_node)
            if ex is None:
                continue

            key = mathy_term_string(variable=ex.variable, exponent=ex.exponent)
            if key == "":
                key = "const"
            if key not in like_counts:
                like_counts[key] = 1
            else:
                like_counts[key] += 1
            if key not in all_indices:
                all_indices[key] = [term_node.r_index]
            else:
                all_indices[key].append(term_node.r_index)

        like_indices: Optional[List[int]] = None
        for key in all_indices.keys():
            if len(all_indices[key]) > 1:
                like_indices = all_indices[key]
        if action_node is not None and touched_term is not None:
            touched_key = mathy_term_string(
                variable=touched_term.variable, exponent=touched_term.exponent
            )
            if touched_key in like_counts and like_counts[touched_key] > 1:
                action_node.all_changed()
                return time_step.termination(features, self.get_win_signal(env_state))

        if env_state.agent.moves_remaining <= 0:
            distances = []
            if like_indices is not None:
                assert action_node is not None and action_node.r_index is not None
                for index in like_indices:
                    distances.append(abs(index - action_node.r_index))
                loss_magnitude = min(distances) / max_index
            else:
                loss_magnitude = 1.0
            lose_signal = EnvRewards.LOSE - loss_magnitude
            return time_step.termination(features, lose_signal)
        return None

    def make_problem(
        self,
        min_terms: int,
        max_terms: int,
        like_terms: int,
        exponent_probability: float,
    ) -> str:
        assert min_terms <= max_terms, "min cannot be greater than max"
        assert like_terms < min_terms, "must have atleast one term that is not like"
        out_terms = []
        total_terms = random.randint(min_terms, max_terms)
        num_diff_terms = total_terms - like_terms
        diff_term_tpls = get_rand_term_templates(
            num_diff_terms + 1, exponent_probability=exponent_probability
        )
        like_term_tpl = diff_term_tpls[-1]
        diff_term_tpls = diff_term_tpls[:-1]

        for i in range(like_terms):
            out_terms.append(like_term_tpl.make())

        for tpl in diff_term_tpls:
            out_terms.append(tpl.make())
        random.shuffle(out_terms)
        problem = " + ".join(out_terms)
        return problem

    def problem_fn(self, params: MathyEnvProblemArgs) -> MathyEnvProblem:
        if params.difficulty == MathyEnvDifficulty.easy:
            text = self.make_problem(
                min_terms=3, max_terms=8, like_terms=2, exponent_probability=0.3
            )
        elif params.difficulty == MathyEnvDifficulty.normal:
            text = self.make_problem(
                min_terms=4, max_terms=7, like_terms=2, exponent_probability=0.5
            )
        elif params.difficulty == MathyEnvDifficulty.hard:
            text = self.make_problem(
                min_terms=5, max_terms=12, like_terms=2, exponent_probability=0.4
            )
        else:
            raise ValueError(f"Unknown difficulty: {params.difficulty}")
        return MathyEnvProblem(text, 2, self.get_env_namespace())



from typing import Optional

import numpy as np
from gym import spaces


class MaskedDiscrete(spaces.Discrete):
    r"""A masked discrete space in :math:`\{ 0, 1, \\dots, n-1 \}`.
    Example::
        >>> MaskedDiscrete(3, mask=(1,1,0))
    """

    def update_mask(self, mask: np.ndarray) -> None:
        assert isinstance(mask, (tuple, list, np.ndarray))
        assert len(mask) == self.n
        self.mask = np.array(mask)

    def __init__(self, n: int, mask: np.ndarray):
        super(MaskedDiscrete, self).__init__(n)  # type:ignore
        self.update_mask(mask)

    def sample(self, mask: Optional[np.ndarray] = None) -> int:
        mask = self.mask if mask is None else mask
        probability = self.mask / np.sum(self.mask)
        return self.np_random.choice(self.n, p=probability)



# coding=utf-8
# Copyright 2018 The TF-Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Changes:
#
#  - 2019/12/09 JD: inline and remove tensorflow dependency/shape support
#  - 2019/12/11 JD: add MathyObservation type hints
#  - 2020/04/04 JD: remove more unused code
"""TimeStep representing a step in the environment.

This file is a mostly direct copy of the implementation from the
[tf_agents](https://github.com/tensorflow/agents) library but has
the dependency on tensorflow removed along with advanced shape
features.

Mathy doesn't use these features and the overhead of loading tensorflow
to pass environment states around is not great for things like CLI start
times.
"""
from typing import NamedTuple

import numpy

from mathy_envs.state import MathyObservation


class TimeStep(NamedTuple):
    step_type: numpy.ndarray
    reward: float
    discount: float
    observation: MathyObservation

    def __hash__(self) -> int:
        return hash(tuple(self))


class StepType(object):
    """Defines the status of a `TimeStep` within a sequence."""

    # Denotes the first `TimeStep` in a sequence.
    FIRST = numpy.array([0], dtype="int32")
    # Denotes any `TimeStep` in a sequence that is not FIRST or LAST.
    MID = numpy.array([1], dtype="int32")
    # Denotes the last `TimeStep` in a sequence.
    LAST = numpy.array([2], dtype="int32")


def transition(
    observation: MathyObservation, reward: float, discount: float = 1.0
) -> TimeStep:
    """Returns a `TimeStep` with `step_type` set equal to `StepType.MID`."""
    return TimeStep(StepType.MID, reward, discount, observation)


def termination(observation: MathyObservation, reward: float) -> TimeStep:
    """Returns a `TimeStep` with `step_type` set to `StepType.LAST`."""
    return TimeStep(StepType.LAST, reward, 00, observation)


def is_terminal_transition(transition: TimeStep) -> bool:
    return bool(transition.step_type == StepType.LAST)



from typing import Any, Dict, Optional, Tuple, Type, Union

import gym
import numpy as np
from gym import error as gym_error
from gym import spaces
from gym.envs.registration import register
from mathy_core.rule import ExpressionChangeRule

from ..env import MathyEnv
from ..state import MathyEnvState
from ..time_step import is_terminal_transition
from ..types import ActionType, MathyEnvProblemArgs
from .masked_discrete import MaskedDiscrete


class MathyGymEnv(gym.Env):
    """A small wrapper around Mathy envs to allow them to work with OpenAI Gym. The
    agents currently use this env wrapper, but it could be dropped in the future."""

    mathy: MathyEnv
    state: Optional[MathyEnvState]
    action_space: MaskedDiscrete
    _challenge: Optional[MathyEnvState]
    env_class: Type[MathyEnv]
    env_problem_args: Optional[MathyEnvProblemArgs]
    mask_as_probabilities: bool

    def __init__(
        self,
        env_class: Type[MathyEnv] = MathyEnv,
        env_problem_args: Optional[MathyEnvProblemArgs] = None,
        env_problem: Optional[str] = None,
        env_max_moves: int = 64,
        mask_as_probabilities: bool = False,
        repeat_problem: bool = False,
        **env_kwargs: Any,
    ):
        self.state = None
        self.mask_as_probabilities = mask_as_probabilities
        self.repeat_problem = repeat_problem
        self.mathy = env_class(**env_kwargs)
        self.env_class = env_class
        self.env_problem_args = env_problem_args
        if env_problem is not None:
            self._challenge = MathyEnvState(
                problem=env_problem, max_moves=env_max_moves
            )
        else:
            self._challenge, _ = self.mathy.get_initial_state(env_problem_args)

        self.action_space = MaskedDiscrete(
            self.action_size, np.array([1] * self.action_size)
        )
        mask = len(self.mathy.rules) * self.mathy.max_seq_len
        values = self.mathy.max_seq_len
        nodes = self.mathy.max_seq_len
        type = 4
        time = 1
        obs_size = mask + values + nodes + type + time
        self.observation_space = spaces.Box(
            low=0, high=1, shape=(obs_size,), dtype=float
        )

    @property
    def action_size(self) -> int:
        return self.mathy.action_size

    def step(
        self, action: Union[int, ActionType]
    ) -> Tuple[np.ndarray, Any, bool, Dict[str, object]]:
        assert self.state is not None, "call reset() before stepping the environment"
        self.state, transition, change = self.mathy.get_next_state(self.state, action)
        done = is_terminal_transition(transition)
        info = {
            "transition": transition,
            "done": done,
            "valid": change.result is not None,
        }
        if done:
            info["win"] = transition.reward > 0.0
        return self._observe(self.state), transition.reward, done, info

    def _observe(self, state: MathyEnvState) -> np.ndarray:
        """Observe the environment at the given state, updating the observation
        space and action space for the given state."""
        action_mask = self.mathy.get_valid_moves(state)
        observation = self.mathy.state_to_observation(
            state, max_seq_len=self.mathy.max_seq_len
        )
        flat_mask = np.array(action_mask).reshape(-1)
        self.action_space.update_mask(flat_mask)
        if self.mask_as_probabilities:
            mask_sum = np.sum(flat_mask)
            if mask_sum > 0.0:
                flat_mask = flat_mask / mask_sum
        nodes = np.array(observation.nodes, dtype="float32").reshape(-1)
        values = np.array(observation.values, dtype="float32").reshape(-1)
        np_observation = np.concatenate(
            np.array(
                [observation.type, observation.time, nodes, values, flat_mask],
                dtype="object",
            ),
            axis=-1,
        )
        return np_observation

    def reset(
        self,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[Dict[Any, Any]] = None,
    ) -> Union[Any, Tuple[Any, Dict[Any, Any]]]:
        if self.state is not None:
            self.mathy.finalize_state(self.state)
        if self.repeat_problem:
            assert self._challenge is not None
            self.state = MathyEnvState.copy(self._challenge)
        else:
            self.state, self.problem = self.mathy.get_initial_state(
                self.env_problem_args
            )
        return self._observe(self.state)

    def reset_with_input(self, problem_text: str, max_moves: int = 16) -> np.ndarray:
        # If the episode is being reset because it ended, assert the validity
        # of the last problem outcome
        if self.state is not None:
            self.mathy.finalize_state(self.state)
        self.reset()
        self.state = MathyEnvState(problem=problem_text, max_moves=max_moves)
        return self._observe(self.state)

    def render(
        self,
        last_action: ActionType = (-1, -1),
        last_reward: float = 0.0,
        last_change: Optional[ExpressionChangeRule] = None,
    ) -> None:
        assert self.state is not None, "call reset() before rendering the env"
        action_name = "initial"
        token_index = -1
        if last_action != (-1, -1):
            action_index, token_index = last_action
            action_name = self.mathy.rules[action_index].name
        else:
            print(f"Problem: {self.state.agent.problem}")
        self.mathy.print_state(
            self.state,
            action_name[:25].lower(),
            token_index,
            change=last_change,
            change_reward=last_reward,
        )


def safe_register(id: str, **kwargs: Any) -> None:
    """Ignore re-register errors."""
    try:
        register(id, **kwargs)
    except gym_error.Error:
        pass



import sys
import typing
from dataclasses import dataclass
from enum import Enum
from typing import List, NamedTuple, Tuple

# Use typing_extensions for Python < 3.8
if sys.version_info < (3, 8):
    from typing_extensions import Literal  # noqa
else:
    from typing import Literal  # noqa


class MathyEnvDifficulty(Enum):
    # The simplest form of problems that demonstrate a task
    easy = "easy"
    # Challenging problems that involve more terms
    normal = "normal"
    # Difficult problems that have intentionally large expressions
    hard = "hard"


MathyEnvDifficultyValue = typing.Literal["easy", "normal", "hard"]

assert set(typing.get_args(MathyEnvDifficultyValue)) == {
    member.value for member in MathyEnvDifficulty
}


@dataclass
class MathyEnvProblemArgs:
    difficulty: MathyEnvDifficulty = MathyEnvDifficulty.easy


class MathyEnvProblem(NamedTuple):
    """Summarize an environment-specific problem that was generated with
    a tuple of (text, complexity, type) where:
     - "text" is the text content of the generated problem
     - "complexity" is an integer value that represents the number of
       terms in the problem text.
     - "type" is a dot namespaced string, e.g. "mathy.poly.simplify"
    """

    text: str
    complexity: int
    type: str


class EnvRewards:
    LOSE = -1.0
    WIN = 1.0
    HELPFUL_MOVE = 0.01
    UNHELPFUL_MOVE = -0.01
    TIMESTEP = -0.01
    PREVIOUS_LOCATION = -0.02
    INVALID_MOVE = -0.5


ActionType = Tuple[int, int]
ActionList = List[ActionType]
RewardList = List[float]
ValueList = List[float]



import random
from itertools import groupby
from typing import Any, Dict, List, Optional, Tuple, Type, Union

import numpy as np
from mathy_core import STOP
from mathy_core.expressions import MathExpression
from mathy_core.parser import ExpressionParser, InvalidSyntax
from mathy_core.rule import BaseRule, ExpressionChangeRule
from mathy_core.rules import (
    AssociativeSwapRule,
    CommutativeSwapRule,
    ConstantsSimplifyRule,
    DistributiveFactorOutRule,
    DistributiveMultiplyRule,
    VariableMultiplyRule,
)
from mathy_core.tree import BinaryTreeNode, VisitStop
from mathy_core.util import compare_expression_string_values, raise_with_history

from . import time_step
from .state import MathyEnvState, MathyEnvStateStep, MathyObservation
from .time_step import is_terminal_transition
from .types import ActionType, EnvRewards, Literal, MathyEnvProblem, MathyEnvProblemArgs

InvalidActionResponses = Literal["raise", "penalize", "terminal"]

INVALID_ACTION_RESPONSES: List[InvalidActionResponses] = [
    "raise",
    "penalize",
    "terminal",
]


class MathyEnv:
    """Implement a math solving game where a player wins by executing the
    right sequence of actions to reduce a math expression to an agreeable
    basic representation in as few moves as possible."""

    rules: List[BaseRule]
    max_moves: int
    max_seq_len: int
    verbose: bool
    reward_discount: float
    parser: ExpressionParser
    valid_actions_mask_cache: Dict[str, List[List[int]]]
    valid_rules_cache: Dict[str, List[int]]
    invalid_action_response: InvalidActionResponses
    previous_state_penalty: bool
    preferred_term_commute: bool

    def __init__(
        self,
        *,
        rules: Optional[List[BaseRule]] = None,
        max_moves: int = 20,
        verbose: bool = False,
        invalid_action_response: InvalidActionResponses = "raise",
        reward_discount: float = 0.99,
        max_seq_len: int = 128,
        previous_state_penalty: bool = True,
        preferred_term_commute: bool = False,
    ):
        self.discount = reward_discount
        self.previous_state_penalty = previous_state_penalty
        self.verbose = verbose
        self.max_moves = max_moves
        self.max_seq_len = max_seq_len
        self.invalid_action_response = invalid_action_response
        self.parser = ExpressionParser()
        if rules is None:
            self.rules = MathyEnv.core_rules(
                preferred_term_commute=preferred_term_commute
            )
        else:
            self.rules = rules
        self.valid_actions_mask_cache = dict()
        self.valid_rules_cache = dict()

        if self.invalid_action_response not in INVALID_ACTION_RESPONSES:
            raise ValueError(
                f"Unknown invalid action behavior: {self.invalid_action_response}\n"
                f"Expected one of: {', '.join(INVALID_ACTION_RESPONSES)}"
            )

    @classmethod
    def core_rules(cls, preferred_term_commute: bool = False) -> List[BaseRule]:
        """Return the mathy core agent actions"""
        return [
            ConstantsSimplifyRule(),
            CommutativeSwapRule(preferred=preferred_term_commute),
            DistributiveMultiplyRule(),
            DistributiveFactorOutRule(),
            AssociativeSwapRule(),
            VariableMultiplyRule(),
        ]

    @property
    def action_size(self) -> int:
        """Return the number of available actions"""
        return len(self.rules) * self.max_seq_len

    def finalize_state(self, state: MathyEnvState) -> None:
        """Perform final checks on a problem state, to ensure the episode yielded
        results that were uncorrupted by transformation errors."""
        from_timestep: MathyEnvStateStep = state.agent.history[0]
        to_timestep: MathyEnvStateStep = state.agent.history[-1]
        compare_expression_string_values(
            str(from_timestep.raw), str(to_timestep.raw), state.agent.history
        )

    def get_env_namespace(self) -> str:
        """Return a unique dot namespaced string representing the current
        environment. e.g. mycompany.envs.differentiate"""
        raise NotImplementedError("subclass must implement this")

    def get_rewarding_actions(self, state: MathyEnvState) -> List[Type[BaseRule]]:
        """Get the list of rewarding action types. When these actions
        are selected, the agent gets a positive reward."""
        # NOTE: by default we give a positive reward for most actions taken. Reward
        #       values are only applied AFTER penalties, so things like reentrant
        #       states become negative reward even if their action is otherwise
        #       rewarding.
        return [
            ConstantsSimplifyRule,
            DistributiveFactorOutRule,
            VariableMultiplyRule,
        ]

    def get_penalizing_actions(self, state: MathyEnvState) -> List[Type[BaseRule]]:
        """Get the list of penalizing action types. When these actions
        are selected, the agent gets a negative reward."""
        return [
            AssociativeSwapRule,
            DistributiveMultiplyRule,
        ]

    def max_moves_fn(
        self, problem: MathyEnvProblem, config: MathyEnvProblemArgs
    ) -> int:
        """Return the environment specific maximum move count for a given prolem."""
        return problem.complexity * 3

    def transition_fn(
        self,
        env_state: MathyEnvState,
        expression: MathExpression,
        features: MathyObservation,
    ) -> Optional[time_step.TimeStep]:
        """Provide environment-specific transitions per timestep."""
        return None

    def problem_fn(self, params: MathyEnvProblemArgs) -> MathyEnvProblem:
        """Return a problem for the environment given a set of parameters
        to control problem generation.

        This is implemented per environment so each environment can
        generate its own dataset with no required configuration."""
        raise NotImplementedError("This must be implemented in a subclass")

    def state_to_observation(
        self, state: MathyEnvState, max_seq_len: Optional[int] = None
    ) -> MathyObservation:
        """Convert an environment state into an observation that can be used
        by a training agent."""

        action_mask = self.get_valid_moves(state)
        observation = state.to_observation(
            move_mask=action_mask, parser=self.parser, max_seq_len=max_seq_len
        )
        return observation

    def get_win_signal(self, env_state: MathyEnvState) -> float:
        """Calculate the reward value for completing the episode. This is done
        so that the reward signal can be scaled based on the time it took to
        complete the episode."""
        tiny = 3e-10
        total_moves = max(tiny, env_state.max_moves)
        # guard against divide by zero with max and a small value
        current_move = max(tiny, total_moves - env_state.agent.moves_remaining)
        bonus = (total_moves / current_move) / total_moves
        # If the episode is not very short, and the agent completes in half
        # the number of allowed steps, double the bonus signal
        if total_moves > 10 and current_move < total_moves / 2:
            bonus *= 2
        return min(2.0, EnvRewards.WIN + bonus)

    def get_lose_signal(self, env_state: MathyEnvState) -> float:
        """Calculate the reward value for failing to complete the episode. This is done
        so that the reward signal can be problem-type dependent."""
        return EnvRewards.LOSE

    def get_state_transition(self, env_state: MathyEnvState) -> time_step.TimeStep:
        """Given an input state calculate the transition value of the timestep.

        # Parameters
            env_state: current env_state

        # Returns
            transition: the current state value transition
        """
        agent = env_state.agent
        expression = self.parser.parse(agent.problem)
        features = env_state.to_observation(
            self.get_valid_moves(env_state), parser=self.parser
        )
        root = expression.get_root()
        assert isinstance(root, MathExpression)

        # Subclass specific win conditions happen here. Custom win-conditions
        # outside of that can override this method entirely.
        result = self.transition_fn(env_state, root, features)
        if result is not None:
            return result

        # Check the turn count last because if the previous move that incremented
        # the turn over the count resulted in a win-condition, we want it to be honored.
        if env_state.agent.moves_remaining <= 0:
            return time_step.termination(features, self.get_lose_signal(env_state))

        # The agent is penalized for returning to a previous state.
        if self.previous_state_penalty is True:
            for key, group in groupby(
                sorted([f"{h.raw}" for h in env_state.agent.history])
            ):
                list_count = len(list(group))
                if list_count <= 1 or key != expression.raw:
                    continue

                # After more than (n) visits to the same state, you lose.
                if list_count > 3:
                    return time_step.termination(
                        features, self.get_lose_signal(env_state)
                    )

                # NOTE: the reward is scaled by # of times this state has been visited
                return time_step.transition(
                    features,
                    reward=EnvRewards.PREVIOUS_LOCATION * list_count,
                    discount=self.discount,
                )

        if len(agent.history) > 0:
            last_timestep = agent.history[-1]
            rule = self.get_rule_from_timestep(last_timestep)
            reward_actions = self.get_rewarding_actions(env_state)
            # The rewarding_actions can be user specified
            for rewarding_class in reward_actions:
                if isinstance(rule, rewarding_class):
                    return time_step.transition(
                        features,
                        reward=EnvRewards.HELPFUL_MOVE,
                        discount=self.discount,
                    )

            penalty_actions = self.get_penalizing_actions(env_state)
            # The rewarding_actions can be user specified
            for penalty_class in penalty_actions:
                if isinstance(rule, penalty_class):
                    return time_step.transition(
                        features,
                        reward=EnvRewards.UNHELPFUL_MOVE,
                        discount=self.discount,
                    )

        # We're in a new state, and the agent is a little older.
        return time_step.transition(
            features, reward=EnvRewards.TIMESTEP, discount=self.discount
        )

    def get_next_state(
        self, env_state: MathyEnvState, action: Union[int, ActionType]
    ) -> Tuple[MathyEnvState, time_step.TimeStep, ExpressionChangeRule]:
        """
        # Parameters
        env_state: current env_state
        action:    a tuple of two integers representing the rule and node to act on

        # Returns
        next_state: env_state after applying action

        transition: the timestep that represents the state transition

        change: the change descriptor describing the change that happened
        """
        action = self.to_action(action)
        agent = env_state.agent
        expression = self.parser.parse(agent.problem)
        assert isinstance(
            action, (tuple, list)
        ), f"Expected tuple action, but received: {type(action)} {action}"
        action_index, token_index = action
        token = self.get_token_at_index(expression, token_index)
        operation = self.rules[action_index]

        op_not_rule = not isinstance(operation, BaseRule)
        op_cannot_apply = token is None or operation.can_apply_to(token) is False
        if token is None or op_not_rule or op_cannot_apply:
            if self.invalid_action_response == "raise":
                steps = int(env_state.max_moves - agent.moves_remaining)
                msg = "Step: {} - Invalid action({}) '{}' for expression '{}'.".format(
                    steps, action, type(operation), expression
                )
                raise_with_history("Invalid Action", msg, agent.history)
            elif self.invalid_action_response == "penalize":
                #
                out_env = env_state.get_out_state(
                    problem=env_state.agent.problem,
                    action=action,
                    moves_remaining=agent.moves_remaining - 1,
                )
                obs = out_env.to_observation(
                    self.get_valid_moves(out_env), parser=self.parser
                )
                transition = time_step.transition(obs, EnvRewards.INVALID_MOVE)
                return out_env, transition, ExpressionChangeRule(BaseRule())
            elif self.invalid_action_response == "terminal":
                out_env = env_state.get_out_state(
                    problem=env_state.agent.problem,
                    action=action,
                    moves_remaining=0,
                )
                obs = out_env.to_observation(
                    self.get_valid_moves(out_env), parser=self.parser
                )
                transition = time_step.termination(obs, self.get_lose_signal(env_state))
                return out_env, transition, ExpressionChangeRule(BaseRule())

        assert token is not None
        change = operation.apply_to(token.clone_from_root())
        assert change.result is not None
        root = change.result.get_root()
        change_name = operation.name
        out_problem = str(root)
        out_env = env_state.get_out_state(
            problem=out_problem,
            action=action,
            moves_remaining=agent.moves_remaining - 1,
        )

        transition = self.get_state_transition(out_env)
        if self.verbose:
            token_idx = int("{}".format(token_index).zfill(3))
            self.print_state(
                out_env, change_name[:25].lower(), token_idx, change, transition.reward
            )
        return out_env, transition, change

    def print_state(
        self,
        env_state: MathyEnvState,
        action_name: str,
        token_index: int = -1,
        change: Optional[ExpressionChangeRule] = None,
        change_reward: float = 0.0,
        pretty: bool = False,
    ) -> None:
        """Render the given state to stdout for visualization"""
        print(
            self.render_state(
                env_state, action_name, token_index, change, change_reward, pretty
            )
        )

    def is_terminal_state(self, env_state: MathyEnvState) -> bool:
        """Determine if a given state is terminal or not.

        # Arguments
        env_state (MathyEnvState): The state to inspect

        # Returns
        (bool): A boolean indicating if the state is terminal or not.
        """
        return is_terminal_transition(self.get_state_transition(env_state))

    def print_history(self, env_state: MathyEnvState, pretty: bool = True) -> None:
        """Render the history of an episode from a given state.

        # Arguments
        env_state (MathyEnvState): The state to render the history of.
        """
        history: List[MathyEnvStateStep] = env_state.agent.history[:]
        initial_step: MathyEnvStateStep = history.pop(0)
        curr_state: MathyEnvState = MathyEnvState(
            problem=initial_step.raw,
            max_moves=env_state.max_moves,
        )
        self.print_state(curr_state, "initial-state", pretty=pretty)
        while len(history) > 0:
            step: MathyEnvStateStep = history.pop(0)
            curr_state, transition, change = self.get_next_state(
                curr_state, step.action
            )
            rule_idx, token_idx = step.action
            rule: BaseRule = self.rules[rule_idx]
            rule_name: str = rule.name[:25].lower()
            self.print_state(
                pretty=pretty,
                env_state=curr_state,
                action_name=rule_name,
                token_index=int(f"{token_idx}".zfill(3)),
                change=change,
                change_reward=transition.reward,
            )

    def render_state(
        self,
        env_state: MathyEnvState,
        action_name: str,
        token_index: int = -1,
        change: Optional[ExpressionChangeRule] = None,
        change_reward: float = 0.0,
        pretty: bool = False,
    ) -> str:
        """Render the given state to a string suitable for printing to a log"""
        changed_problem = env_state.agent.problem
        if change is not None and change.result is not None:
            root = change.result.get_root()
            assert isinstance(root, MathExpression)
            changed_problem = root.terminal_text

        action_name = f"{action_name.lower()}({token_index})"
        output = """{:<25} | {}""".format(action_name.lower(), changed_problem)

        def get_move_shortname(index: int, move: int) -> str:
            if move == 0:
                return "--"
            if move >= len(self.rules):
                return "xx"
            return self.rules[index].code.lower()

        moves_left = str(env_state.agent.moves_remaining).zfill(2)
        valid_rules = self.get_valid_rules(env_state)
        valid_moves = self.get_valid_moves(env_state)
        num_moves = "{}".format(len(np.nonzero(np.array(valid_moves))[0])).zfill(3)
        move_codes = [get_move_shortname(i, m) for i, m in enumerate(valid_rules)]
        moves = " ".join(move_codes)
        reward = f"{change_reward:.2}"
        reward = f"{reward:<5}"
        if pretty:
            return output
        return f"{num_moves} | {moves} | {moves_left} | {reward} | {output}"

    def random_action(
        self,
        expression: MathExpression,
        rule: Optional[Type[BaseRule]] = None,
    ) -> Tuple[int, int]:
        """Get a random action index that represents a particular rule"""

        if rule is not None:
            found = -1
            for rule_idx, r in enumerate(self.rules):
                if isinstance(r, rule):  # type:ignore
                    found = rule_idx
                    break
            if found == -1:
                raise ValueError(
                    "The action {rule} does not exist in the environment rule list"
                )
            all_actions = self.get_actions_for_node(expression, [rule])
            valid_actions = np.nonzero(all_actions[found])[0].tolist()
            action: int = random.choice(valid_actions)
            return (found, int(action))

        all_actions = self.get_actions_for_node(expression)
        valid_rules = [i for i, r in enumerate(all_actions) if 1 in r]
        if len(valid_rules) == 0:
            raise ValueError(f"no valid actions for expression: {expression}")
        chosen_rule = random.choice(valid_rules)
        valid_actions = [i for i, r in enumerate(all_actions[chosen_rule]) if r == 1]
        action = random.choice(valid_actions)
        return chosen_rule, action

    def get_initial_state(
        self, params: Optional[MathyEnvProblemArgs] = None, print_problem: bool = True
    ) -> Tuple[MathyEnvState, MathyEnvProblem]:
        """Generate an initial MathyEnvState for an episode"""
        config = params if params is not None else MathyEnvProblemArgs()
        prob: MathyEnvProblem = self.problem_fn(config)
        self.valid_actions_mask_cache = dict()
        self.valid_rules_cache = dict()
        self.parser.clear_cache()
        self.max_moves = self.max_moves_fn(prob, config)

        # Build and return the initial state
        env_state = MathyEnvState(
            problem=prob.text,
            problem_type=self.get_env_namespace(),
            max_moves=self.max_moves,
            num_rules=len(self.rules),
        )
        if print_problem and self.verbose:
            self.print_state(env_state, "initial-state")
        return env_state, prob

    def get_agent_actions_count(self, env_state: MathyEnvState) -> int:
        """Return number of all possible actions"""
        node_count = len(self.parser.parse(env_state.agent.problem).to_list())
        return self.action_size * node_count

    def get_token_at_index(
        self, expression: MathExpression, index: int
    ) -> Optional[MathExpression]:
        """Get the token that is `index` from the left of the expression"""
        count = 0
        result: Optional[MathExpression] = None

        def visit_fn(
            node: BinaryTreeNode, depth: int, data: Any
        ) -> Optional[VisitStop]:
            nonlocal result, count
            result = node  # type:ignore
            if count == index:
                return STOP
            count = count + 1
            return None

        expression.visit_inorder(visit_fn)
        return result

    def get_valid_moves(self, env_state: MathyEnvState) -> List[List[int]]:
        """Get a 2d list describing the valid moves for the current state.

        The first dimension contains the list of known rules in the order that
        they're registered, and the second dimension contains a list of the max
        sequence length size that is 1/0 representing that the node at that index
        for the given rule is valid.
        """
        agent = env_state.agent
        expression: Optional[MathExpression] = None
        try:
            expression = self.parser.parse(agent.problem)
        except InvalidSyntax as err:
            raise_with_history(
                self.get_env_namespace(), err.message, env_state.agent.history
            )
        assert expression is not None
        return self.get_actions_for_node(expression)

    def get_valid_rules(self, env_state: MathyEnvState) -> List[int]:
        """Get a vector the length of the number of valid rules that is
        filled with 0/1 based on whether the rule has any nodes in the
        expression that it can be applied to.

        !!! note

            If you want to get a list of which nodes each rule can be
            applied to, prefer to use the `get_valid_moves` method.
        """
        key = self.to_hash_key(env_state)
        if key in self.valid_rules_cache:
            return self.valid_rules_cache[key]
        expression = self.parser.parse(env_state.agent.problem)
        actions = [0] * len(self.rules)
        for rule_index, rule in enumerate(self.rules):
            nodes = rule.find_nodes(expression)
            actions[rule_index] = 0 if len(nodes) == 0 else 1
        self.valid_rules_cache[key] = actions[:]
        return actions

    def get_rule_from_timestep(self, time_step: MathyEnvStateStep) -> BaseRule:
        return self.rules[time_step.action[0]]

    def get_actions_for_node(
        self,
        expression: MathExpression,
        rule_list: Optional[List[Type[BaseRule]]] = None,
    ) -> List[List[int]]:
        """Return a valid actions mask for the given expression and rule list.

        Action masks are 2d lists of length (num_rules, max_seq_len) where a 0 indicates
        the action is not valid in the current state, and a 1 indicates that it is
        a valid action to take."""
        key = str(expression)
        if rule_list is None and key in self.valid_actions_mask_cache:
            return self.valid_actions_mask_cache[key][:]
        node_count = len(expression.to_list())
        rule_count = len(self.rules)
        actions = [[0] * node_count for _ in range(rule_count)]
        for rule_index, rule in enumerate(self.rules):
            if rule_list is not None:
                if not isinstance(rule, tuple(rule_list)):  # type:ignore
                    continue
            nodes = rule.find_nodes(expression)
            for node in nodes:
                assert node.r_index is not None
                actions[rule_index][node.r_index] = 1
        if rule_list is None:
            self.valid_actions_mask_cache[key] = actions[:]
        return actions

    def to_hash_key(self, env_state: MathyEnvState) -> str:
        """Convert env_state to a string for MCTS cache"""
        return env_state.agent.problem

    def to_action(self, action: Union[int, ActionType]) -> ActionType:
        """Resolve a given action input to a tuple of (rule_index, node_index).

        When given an int, it is treated as an index into the flattened 2d action
        space. When given a tuple, it is assumed to be (rule, node)"""
        if isinstance(action, (tuple, list)):
            return action
        token_index = action % self.max_seq_len
        action_index = int((action - token_index) / self.max_seq_len)
        return action_index, token_index



import random
from typing import Any

import pytest
from mathy_core.rules import AssociativeSwapRule, CommutativeSwapRule

from mathy_envs import MathyEnv, MathyEnvState
from mathy_envs.env import INVALID_ACTION_RESPONSES
from mathy_envs.envs.poly_simplify import PolySimplify
from mathy_envs.time_step import is_terminal_transition
from mathy_envs.types import EnvRewards


def test_env_init():
    env = MathyEnv()
    assert env is not None
    # Default env is abstract and cannot be directly used for problem solving
    with pytest.raises(NotImplementedError):
        env.get_initial_state()
    with pytest.raises(NotImplementedError):
        env.get_env_namespace()


def test_env_init_check_invalid_action_response():
    with pytest.raises(ValueError):
        MathyEnv(invalid_action_response="something_wrong")  # type:ignore
    option: Any
    for option in INVALID_ACTION_RESPONSES:
        assert MathyEnv(invalid_action_response=option) is not None


def test_env_action_masks():
    problem = "4x + 2x"
    env = MathyEnv(invalid_action_response="raise")
    env_state = MathyEnvState(problem=problem, max_moves=35)
    valid_mask = env.get_valid_moves(env_state)
    assert len(valid_mask) == len(env.rules)
    assert len(valid_mask[0]) == len(env.parser.parse(problem).to_list())


def test_env_random_actions():
    env = MathyEnv(invalid_action_response="raise")
    state = MathyEnvState(problem="4x + 2x + 7 + y")
    expression = env.parser.parse(state.agent.problem)
    # Can select random actions of the given type
    action = env.random_action(expression, AssociativeSwapRule)
    env.get_next_state(state, action)

    # Can select random actions from all types
    state = MathyEnvState(problem="4x + 2x + 7 + y")
    expression = env.parser.parse(state.agent.problem)
    action = env.random_action(expression)
    env.get_next_state(state, action)


def test_env_invalid_action_behaviors():

    problem = "4x + 2x"
    env = MathyEnv(invalid_action_response="raise")
    env_state = MathyEnvState(problem=problem, max_moves=35)
    rule_actions = env.get_valid_moves(env_state)
    rule_indices = [i for i, value in enumerate(rule_actions) if 1 not in value]
    random.shuffle(rule_indices)
    rule_nodes = rule_actions[rule_indices[0]]
    node_indices = [i for i, value in enumerate(rule_nodes) if value == 0]
    action = (rule_indices[0], node_indices[0])

    # Raise an error when selecting an invalid action
    env_state = MathyEnvState(problem=problem, max_moves=35)
    with pytest.raises(ValueError):
        env.get_next_state(env_state, action)

    # Penalize the agent for choosing an invalid action
    env = MathyEnv(invalid_action_response="penalize")
    env_state = MathyEnvState(problem=problem, max_moves=35)
    _, transition, _ = env.get_next_state(env_state, action)
    assert transition.reward == EnvRewards.INVALID_MOVE
    assert is_terminal_transition(transition) is False

    # End the episode when choosing an invalid action
    env = MathyEnv(invalid_action_response="terminal")
    env_state = MathyEnvState(problem=problem, max_moves=35)
    _, transition, _ = env.get_next_state(env_state, action)
    # a transition is returned with error_invalid=False
    assert is_terminal_transition(transition) is True


def test_env_terminal_conditions():

    expectations = [
        ("70656 * (x^2 * z^6)", True),
        ("b * (44b^2)", False),
        ("z * (1274z^2)", False),
        ("4x^2", True),
        ("100y * x + 2", True),
        ("10y * 10x + 2", False),
        ("10y + 1000y * (y * z)", False),
        ("4 * (5y + 2)", False),
        ("2", True),
        ("4x * 2", False),
        ("4x * 2x", False),
        ("4x + 2x", False),
        ("4 + 2", False),
        ("3x + 2y + 7", True),
        ("3x^2 + 2x + 7", True),
        ("3x^2 + 2x^2 + 7", False),
    ]

    # Valid solutions but out of scope so they aren't counted as wins.
    #
    # This works because the problem sets exclude this type of > 2 term
    # polynomial expressions
    out_of_scope_valid = []

    env = PolySimplify()
    for text, is_win in expectations + out_of_scope_valid:
        env_state = MathyEnvState(problem=text)
        reward = env.get_state_transition(env_state)
        assert text == text and env.is_terminal_state(env_state) == bool(is_win)
        assert text == text and is_terminal_transition(reward) == bool(is_win)


@pytest.mark.parametrize("pretty", [True, False])
def test_env_print_history(pretty: bool):
    env = PolySimplify()
    env_state = MathyEnvState(problem="4+2")
    for i in range(10):
        env_state = env_state.get_out_state(
            problem="4+2" if i % 2 == 0 else "2+4",
            moves_remaining=10 - i,
            action=(1, 1),
        )
    env.print_history(env_state, pretty=pretty)


def test_env_finalize_state():
    env = PolySimplify()

    env_state = MathyEnvState(problem="4x + 2x").get_out_state(
        problem="1337", action=(1, 1), moves_remaining=0
    )
    with pytest.raises(ValueError):
        env.finalize_state(env_state)

    env_state = MathyEnvState(problem="4x + 2x").get_out_state(
        problem="4x + 2", action=(1, 1), moves_remaining=0
    )
    with pytest.raises(ValueError):
        env.finalize_state(env_state)

    env_state = MathyEnvState(problem="4x + 2x").get_out_state(
        problem="4x + 2y", action=(1, 1), moves_remaining=0
    )
    with pytest.raises(ValueError):
        env.finalize_state(env_state)


def test_mathy_env_can_timestep_loop():
    env = MathyEnv()
    assert env is not None
    problem = "5y * 9x + 8z + 8x + 3z * 10y * 11x + 10y"
    env_state = MathyEnvState(problem=problem, max_moves=35)
    for i in range(3):
        rule_actions = env.get_valid_moves(env_state)
        rule_indices = [i for i, value in enumerate(rule_actions) if 1 in value]
        random.shuffle(rule_indices)
        rule_nodes = rule_actions[rule_indices[0]]
        node_indices = [i for i, value in enumerate(rule_nodes) if value == 1]
        env_state, value, changed = env.get_next_state(
            env_state, (rule_indices[0], node_indices[0])
        )
    assert env_state.to_observation([]) is not None


def test_mathy_env_invalid_action_behaviors():
    env = MathyEnv()
    assert env is not None
    problem = "5y * 9x + 8z + 8x + 3z * 10y * 11x + 10y"
    env_state = MathyEnvState(problem=problem, max_moves=35)
    for i in range(3):
        rule_actions = env.get_valid_moves(env_state)
        rule_indices = [i for i, value in enumerate(rule_actions) if 1 in value]
        random.shuffle(rule_indices)
        rule_nodes = rule_actions[rule_indices[0]]
        node_indices = [i for i, value in enumerate(rule_nodes) if value == 1]
        env_state, value, changed = env.get_next_state(
            env_state, (rule_indices[0], node_indices[0])
        )
    assert env_state.to_observation([]) is not None


def test_mathy_env_preferred_term_commute():
    rule_idx = 1
    problem = "5y"
    env_state = MathyEnvState(problem=problem, max_moves=1)

    env = MathyEnv(preferred_term_commute=False)
    assert isinstance(env.rules[rule_idx], CommutativeSwapRule), "update rule_idx"
    commute_nodes = env.get_valid_moves(env_state)[rule_idx]
    assert 1 not in commute_nodes, "shouldn't be able to commute preferred order terms"

    env = MathyEnv(preferred_term_commute=True)
    commute_nodes = env.get_valid_moves(env_state)[rule_idx]
    assert 1 in commute_nodes, "should be able to commute preferred order terms"


def test_mathy_env_previous_state_penalty():
    """When previous_state_penalty=True, a negative reward is given when
    revisiting already seen problem states. If an agent revisits the
    state too many times, the game ends."""

    # We define the input problem with 3 nodes for simplicity
    # "x * y" == ["x","*","y"]
    # Because the tree is small and balanced, we can commute the
    # same node over and over to flip back-and-forth between x * y
    # and y * x.
    problem = "x * y"
    env = MathyEnv(previous_state_penalty=True)
    rule_idx = 1
    node_idx = 1
    assert isinstance(env.rules[rule_idx], CommutativeSwapRule), "update rule_idx"
    action = (rule_idx, node_idx)
    env_state = MathyEnvState(problem=problem, max_moves=10)
    # Commute the first time so we are revisit the initial state
    # as we apply the same action again.
    env_state, _, _ = env.get_next_state(env_state, action)

    # After three visits to the same state, the game ends.
    last_penalty = 0.0
    found_terminal = False
    for i in range(3):
        env_state, transition, changed = env.get_next_state(env_state, action)
        assert transition.reward < 0.0
        # The penalty scales up based on the number of visits to the state
        assert transition.reward < last_penalty
        last_penalty = transition.reward

        if i < 2:
            # Visit the opposite state and ignore it (we only care about revisiting
            # the initial state)
            env_state, _, _ = env.get_next_state(env_state, action)
        else:
            # After the third time, we should receive a terminal transition
            assert is_terminal_transition(transition) is True
            found_terminal = True

    assert found_terminal is True, "did not receive expected terminal transition"


def test_mathy_env_win_conditions():

    expectations = [
        ("70656 * (x^2 * z^6)", True),
        ("b * (44b^2)", False),
        ("z * (1274z^2)", False),
        ("4x^2", True),
        ("100y * x + 2", True),
        ("10y * 10x + 2", False),
        ("10y + 1000y * (y * z)", False),
        ("4 * (5y + 2)", False),
        ("2", True),
        ("4x * 2", False),
        ("4x * 2x", False),
        ("4x + 2x", False),
        ("4 + 2", False),
        ("3x + 2y + 7", True),
        ("3x^2 + 2x + 7", True),
        ("3x^2 + 2x^2 + 7", False),
    ]

    # Valid solutions but out of scope so they aren't counted as wins.
    #
    # This works because the problem sets exclude this type of > 2 term
    # polynomial expressions
    out_of_scope_valid = []

    env = PolySimplify()
    for text, is_win in expectations + out_of_scope_valid:
        env_state = MathyEnvState(problem=text)
        reward = env.get_state_transition(env_state)
        assert text == text and is_terminal_transition(reward) == bool(is_win)



try:
    import gym  # noqa
except ImportError:
    raise ImportError(
        """
The "gym" library must be installed to use mathy_envs.gym submodule. Please try:

    pip install mathy_envs[gym]

"""
    )
from .gym_binomial_distribute import *  # noqa
from .gym_complex_simplify import *  # noqa
from .gym_poly_blockers import *  # noqa
from .gym_poly_combine_in_place import *  # noqa
from .gym_poly_commute_like_terms import *  # noqa
from .gym_poly_grouping import *  # noqa
from .gym_poly_haystack_like_terms import *  # noqa
from .gym_poly_simplify import *  # noqa
from .masked_discrete import *  # noqa
from .mathy_gym_env import *  # noqa



from enum import IntEnum
from typing import Any, Dict, List, NamedTuple, Optional
from zlib import adler32

import numpy as np
from mathy_core.expressions import ConstantExpression, MathExpression, MathTypeKeys
from mathy_core.parser import ExpressionParser

from .types import ActionType
from .util import pad_array

PROBLEM_TYPE_HASH_BUCKETS = 128

NodeIntList = List[int]
NodeValuesFloatList = List[float]
NodeMaskIntList = List[List[int]]
ProblemTypeIntList = List[int]
TimeFloatList = List[float]


WindowNodeIntList = List[NodeIntList]
WindowNodeMaskIntList = List[NodeMaskIntList]
WindowNodeValuesFloatList = List[NodeValuesFloatList]
WindowProblemTypeIntList = List[ProblemTypeIntList]
WindowTimeFloatList = List[TimeFloatList]


# Input type for mathy models
MathyInputsType = Dict[str, Any]


class ObservationFeatureIndices(IntEnum):
    nodes = 0
    mask = 1
    values = 2
    type = 3
    time = 4


class MathyObservation(NamedTuple):
    """A featurized observation from an environment state."""

    @classmethod
    def empty(cls, template: "MathyObservation") -> "MathyObservation":
        return MathyObservation(
            nodes=[MathTypeKeys["empty"]],
            values=[0.0],
            mask=template.mask,
            type=[0, 0],
            time=[0.0],
        )

    nodes: NodeIntList
    mask: NodeMaskIntList
    values: NodeValuesFloatList
    type: ProblemTypeIntList
    time: TimeFloatList


# fmt: off
MathyObservation.nodes.__doc__ = "tree node types in the current environment state shape=[n,]" # noqa
MathyObservation.mask.__doc__ = "0/1 mask where 0 indicates an invalid action shape=[n,]" # noqa
MathyObservation.values.__doc__ = "tree node value sequences, with non number indices set to 0.0 shape=[n,]" # noqa
MathyObservation.type.__doc__ = "two column hash of problem environment type shape=[2,]" # noqa
MathyObservation.time.__doc__ = "float value between 0.0 and 1.0 indicating the time elapsed shape=[1,]" # noqa
# fmt: on


class MathyEnvStateStep(NamedTuple):
    """Capture summarized environment state for a previous timestep so the
    agent can use context from its history when making new predictions."""

    raw: str
    action: ActionType


# fmt: off
MathyEnvStateStep.raw.__doc__ = "the input text at the timestep" # noqa
MathyEnvStateStep.action.__doc__ = "a tuple indicating the chosen action and the node it was applied to" # noqa
# fmt: on


_problem_hash_cache: Optional[Dict[str, List[int]]] = None


class MathyAgentState:
    """The state related to an agent for a given environment state"""

    moves_remaining: int
    problem: str
    problem_type: str
    reward: float
    action: ActionType
    history: List[MathyEnvStateStep]

    def __init__(
        self,
        moves_remaining: int,
        problem: str,
        problem_type: str,
        reward: float = 0.0,
        history: Optional[List[MathyEnvStateStep]] = None,
    ):
        self.moves_remaining = moves_remaining
        self.problem = problem
        self.reward = reward
        self.problem_type = problem_type
        self.history = (
            history[:]
            if history is not None
            else [MathyEnvStateStep(problem, (-1, -1))]
        )

    @classmethod
    def copy(cls, from_state: "MathyAgentState") -> "MathyAgentState":
        return MathyAgentState(
            moves_remaining=from_state.moves_remaining,
            problem=from_state.problem,
            reward=from_state.reward,
            problem_type=from_state.problem_type,
            history=from_state.history,
        )


class MathyEnvState(object):
    """Class for holding environment state and extracting features
    to be passed to the policy/value neural network.

    Mutating operations all return a copy of the environment adapter
    with its own state.

    This allocation strategy requires more memory but removes a class
    of potential issues around unintentional sharing of data and mutation
    by two different sources.
    """

    agent: "MathyAgentState"
    max_moves: int
    num_rules: int

    def __init__(
        self,
        state: Optional["MathyEnvState"] = None,
        problem: Optional[str] = None,
        max_moves: int = 10,
        num_rules: int = 0,
        problem_type: str = "mathy.unknown",
    ):
        self.max_moves = max_moves
        self.num_rules = num_rules
        if problem is not None:
            self.agent = MathyAgentState(max_moves, problem, problem_type)
        elif state is not None:
            self.num_rules = state.num_rules
            self.max_moves = state.max_moves
            self.agent = MathyAgentState.copy(state.agent)

    @classmethod
    def copy(cls, from_state: "MathyEnvState") -> "MathyEnvState":
        return MathyEnvState(state=from_state)

    def clone(self) -> "MathyEnvState":
        return MathyEnvState(state=self)

    def get_out_state(
        self, problem: str, action: ActionType, moves_remaining: int
    ) -> "MathyEnvState":
        """Get the next environment state based on the current one with updated
        history and agent information based on an action being taken."""
        out_state = MathyEnvState.copy(self)
        agent = out_state.agent
        agent.history.append(MathyEnvStateStep(problem, action))
        agent.problem = problem
        agent.action = action
        agent.moves_remaining = moves_remaining
        return out_state

    def get_problem_hash(self) -> ProblemTypeIntList:
        """Return a two element array with hashed values for the current environment
        namespace string.

        # Example

        - `mycorp.envs.solve_impossible_problems` -> `[12375561, -12375561]`

        """
        global _problem_hash_cache
        if _problem_hash_cache is None:
            _problem_hash_cache = {}
        if self.agent.problem_type not in _problem_hash_cache:
            type_str = self.agent.problem_type
            # Generate a normalized hash with values between 0.0 and 1.0
            # The adler32 crc is used in zip files and is determinstic
            # across runs while also being fast to calculate. It is NOT
            # cryptographically secure.
            hashes = np.asfarray(
                [
                    adler32(type_str.encode(encoding="utf-8")),
                    adler32(type_str[::-1].encode(encoding="utf-8")),
                    adler32(type_str.upper().encode(encoding="utf-8")),
                    adler32(type_str.replace(".", "").encode(encoding="utf-8")),
                ]
            )
            if hashes.sum() != 0.0:
                hashes = (hashes - min(hashes)) / (max(hashes) - min(hashes))
            values = hashes.tolist()

            _problem_hash_cache[self.agent.problem_type] = values

        return _problem_hash_cache[self.agent.problem_type]

    def to_observation(
        self,
        move_mask: Optional[NodeMaskIntList] = None,
        hash_type: Optional[ProblemTypeIntList] = None,
        parser: Optional[ExpressionParser] = None,
        normalize: bool = True,
        max_seq_len: Optional[int] = None,
    ) -> MathyObservation:
        """Convert a state into an observation"""
        if parser is None:
            parser = ExpressionParser()
        if hash_type is None:
            hash_type = self.get_problem_hash()
        expression = parser.parse(self.agent.problem)
        nodes: List[MathExpression] = expression.to_list()
        vectors: NodeIntList = []
        values: NodeValuesFloatList = []
        if move_mask is None:
            move_mask = np.zeros(len(nodes)).tolist()
        assert move_mask is not None
        for node in nodes:
            vectors.append(node.type_id)
            if isinstance(node, ConstantExpression):
                assert node.value is not None
                values.append(float(node.value))
            else:
                values.append(0.0)

        # The "types" and "values" can be normalized 0-1
        if normalize is True:
            # https://bit.ly/3irAalH
            x = np.asfarray(values)
            if x.sum() != 0.0:
                x = (x - min(x)) / (max(x) - min(x) + 1e-32)
            values = x.tolist()
            x = np.asfarray(vectors)
            if x.sum() != 0.0:
                x = (x - min(x)) / (max(x) - min(x) + 1e-32)
            vectors = x.tolist()

        # Pass a 0-1 value indicating the relative episode time where 0.0 is
        # the episode start, and 1.0 is the episode end as indicated by the
        # maximum allowed number of actions.
        step = int(self.max_moves - self.agent.moves_remaining)
        time = int(step / self.max_moves * 10)

        # Pad observations to max_seq_len if specified
        if max_seq_len is not None:
            values = pad_array(values, max_seq_len, 0.0)
            vectors = pad_array(vectors, max_seq_len, 0.0)
            move_mask = [pad_array(m, max_seq_len, 0) for m in move_mask]

        return MathyObservation(
            nodes=vectors, mask=move_mask, values=values, type=hash_type, time=[time]
        )

    @classmethod
    def from_string(cls, input_string: str) -> "MathyEnvState":
        """Convert a string representation of state into a state object"""
        sep = "@"
        history_sep = ","
        # remove any padding from string inputs (if added by to_np call)
        input_string = input_string.rstrip()
        inputs = input_string.split(sep)
        state = MathyEnvState()
        state.max_moves = int(inputs[0])
        state.num_rules = int(inputs[1])
        state.agent = MathyAgentState(
            moves_remaining=int(inputs[2]),
            problem=str(inputs[3]),
            problem_type=str(inputs[4]),
            reward=float(inputs[5]),
            history=[],
        )
        history = inputs[6:]
        for step in history:
            raw, action, focus = step.split(history_sep)
            state.agent.history.append(
                MathyEnvStateStep(raw, (int(action), int(focus)))
            )
        return state

    def to_string(
        self,
    ) -> str:
        """Convert a state object into a string representation"""
        sep = "@"
        assert self.agent is not None, "invalid state"
        out = [
            str(self.max_moves),
            str(self.num_rules),
            str(self.agent.moves_remaining),
            str(self.agent.problem),
            str(self.agent.problem_type),
            str(self.agent.reward),
        ]
        for step in self.agent.history:
            out.append(
                ",".join([str(step.raw), str(step.action[0]), str(step.action[1])])
            )
        return sep.join(out)

    @classmethod
    def from_np(cls, input_bytes: np.ndarray) -> "MathyEnvState":
        """Convert a numpy object into a state object"""
        input_string = "".join([chr(o) for o in input_bytes.tolist()])
        state = cls.from_string(input_string)
        return state

    def to_np(self, pad_to: Optional[int] = None) -> np.ndarray:
        """Convert a state object into a numpy representation"""
        string = self.to_string()
        if pad_to is not None:
            assert pad_to >= len(string), "input is larger than pad size!"
            to_pad = pad_to - len(string)
            string += " " * to_pad
        return np.array([ord(c) for c in string])
